/*
  *
  * This file is a part of CorePDF.
  * A PDF viewer for C Suite.
  * Copyright 2019 CuboCore Group
  *
  *
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
  * the Free Software Foundation; either version 3 of the License, or
  * (at your option) any later version.
  *
  *
  *
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
  *
  *
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, vsit http://www.gnu.org/licenses/.
  *
  */


#include "corepdf.h"
#include "ui_corepdf.h"
#include "Navigation.hpp"

#include <QListWidgetItem>
#include <QLabel>
#include <QBasicTimer>

#include <cprime/cprime.h>
#include <cprime/pinit.h>
#include <cprime/shareit.h>
#include <cprime/appopenfunc.h>
#include <cprime/messageengine.h>
#include <cprime/pinmanage.h>

#include "settings.h"

#include "qpdfview.h"
#include "Document.hpp"

corepdf::corepdf(QWidget *parent) : QWidget(parent)
	, ui(new Ui::corepdf)
	, smi(new settings)
{
	ui->setupUi(this);

	// Set the transparent color to the main widget
	QPalette p(palette());

	p.setColor(QPalette::Base, Qt::transparent);
	this->setPalette(p);

	loadSettings();
	startsetup();
	setupIcons();
}


corepdf::~corepdf()
{
	delete ui;
	delete PdfDoc;
	delete PdfWidget;
	delete smi;
}


/**
  * @brief Setup ui elements
  */
void corepdf::startsetup()
{
	PdfWidget = new QPdfView(this);
	PdfWidget->setLayoutContinuous(continuousPage);
	PdfWidget->setZoomMode((QPdfView::ZoomMode)zoomMode);
	PdfWidget->setZoomFactor(zoomFactor * 1.0 / 100.0);
	ui->u->addWidget(PdfWidget);

	connect(
		PdfWidget, &QPdfView::zoomFactorChanged, [ = ](qreal zoomFactor) {
		QList<int> zooms = { 10, 20, 50, 75, 100, 150, 200, 400 };
		cur_zoom_idx     = 2 + zooms.indexOf((int)(PdfWidget->zoomFactor() * 100));
		ui->zoomCB->setCurrentText(QString("%1%").arg(zooms.value(cur_zoom_idx - 2)));
	}
		);

	connect(
		PdfWidget, &QPdfView::matchesFound, [ = ](int numMatches) {
		ui->searchCount->setText(QString("%1 match%2").arg(numMatches).arg(numMatches == 1 ? "" : "es"));
	}
		);

	connect(
		PdfWidget, &QPdfView::searchComplete, [ = ](int numMatches) {
		ui->searchCount->setText(QString("%1 match%2").arg(numMatches).arg(numMatches == 1 ? "" : "es"));
	}
		);

	switch (zoomMode)
	{
	case 0:
		ui->zoomCB->setCurrentIndex(0);
		cur_zoom_idx = 0;
		break;

	case 1:
		ui->zoomCB->setCurrentIndex(1);
		cur_zoom_idx = 1;
		break;

	default:
		ui->zoomCB->setCurrentText(QString("%1%").arg(zoomFactor));
		cur_zoom_idx = ui->zoomCB->currentIndex();
		break;
	}

	QMenu *menu = new QMenu(ui->lytBtn);

	menu->addAction(
		QIcon::fromTheme("view-pages-single"), "Single Page", [ = ]() {
			ui->lytBtn->setIcon(QIcon::fromTheme("view-pages-single"));
			ui->lytBtn->setText("Single Page");
			PdfWidget->setPageLayout(QPdfView::SinglePage);
		}
	);

	menu->addAction(
		QIcon::fromTheme("view-pages-facing"), "Facing Pages", [ = ]() {
			ui->lytBtn->setIcon(QIcon::fromTheme("view-pages-facing"));
			ui->lytBtn->setText("Facing Pages");
			PdfWidget->setPageLayout(QPdfView::FacingPages);
		}
	);

	menu->addAction(
		QIcon::fromTheme("view-pages-facing-first-centered"), "Book View", [ = ]() {
			ui->lytBtn->setIcon(QIcon::fromTheme("view-pages-facing-first-centered"));
			ui->lytBtn->setText("Book View");
			PdfWidget->setPageLayout(QPdfView::BookView);
		}
	);

	menu->actions()[pageLayout]->trigger();

	menu->addSeparator();
	contLytAct = menu->addAction(QIcon::fromTheme("view-pages-continuous"), "Continuous");
	contLytAct->setCheckable(true);
	contLytAct->setChecked(continuousPage);
	connect(
		contLytAct, &QAction::triggered, [ = ]() {
			PdfWidget->setLayoutContinuous(contLytAct->isChecked());
		}
	);
	ui->lytBtn->setMenu(menu);
	ui->lytBtn->setIconSize(toolsIconSize);

	// all toolbuttons icon size in sideView
	QList<QToolButton *> toolBtns = ui->sideView->findChildren<QToolButton *>();

	for (QToolButton *b: toolBtns){
		if (b){
			b->setIconSize(toolsIconSize);}
	}

	// all toolbuttons icon size in toolBar
	QList<QToolButton *> toolBtns2 = ui->toolBar->findChildren<QToolButton *>();

	for (QToolButton *b: toolBtns2){
		if (b){
			b->setIconSize(toolsIconSize);}
	}

	shotcuts();

	ui->sideView->setVisible(false);
	ui->menu->setEnabled(false);

	ui->appTitle->setAttribute(Qt::WA_TransparentForMouseEvents);
	ui->appTitle->setFocusPolicy(Qt::NoFocus);
	this->resize(800, 500);


	ui->searchB->setVisible(false);

    if (uiMode == 2) {
        // setup mobile UI

        this->setWindowState(Qt::WindowMaximized);

        ui->activitiesList->setVisible(false);
        ui->fileName->setVisible(false);

        ui->activitiesList->setVisible(false);
    } else {
        // setup desktop or tablet UI

        if(windowMaximized){
            this->setWindowState(Qt::WindowMaximized);
            qDebug() << "window is maximized";
        } else{
            this->resize(windowSize);
        }

        if(activities){
            loadActivities();
        }
    }

	ui->page->setCurrentIndex(0);

	// Zoom label
	pageCountLbl = new QLabel(this);
	pageCountLbl->setStyleSheet("QLabel{ margin: 5px; padding: 5px; background-color: rgba(0, 0, 0, 0.5); border-radius: 3px; }");
	pageCountLbl->setText("Page 0 of 0");
	ui->gridLayout->addWidget(pageCountLbl, 0, 0, 0, 0, Qt::AlignRight | Qt::AlignBottom);

	connect(ui->jumpToPage, &QToolButton::clicked, this, &corepdf::jumpPage);

	connect(ui->addFiles, SIGNAL(clicked()), this, SLOT(openFileDialog()));
	connect(ui->open, SIGNAL(clicked()), this, SLOT(openFileDialog()));

	QScroller::grabGesture(PdfWidget, QScroller::LeftMouseButtonGesture);

	connect(PdfWidget->pageNavigation(), &CorePDF::Navigation::currentPageChanged, this, &corepdf::updatePageNumber);

	delayTimer = new QBasicTimer();

	connect(ui->searchHere, &QLineEdit::textEdited, [ = ]() {
		delayTimer->start(500, Qt::PreciseTimer, this);
	});

	connect(ui->closeSearch, &QToolButton::clicked, [ = ]() {
		ui->searchHere->clear();
		PdfWidget->searchText(QString());
	});
}


/**
  * @brief Loads application settings
  */
void corepdf::loadSettings()
{
    // get CSuite's settings
    toolsIconSize = smi->getValue("CoreApps", "ToolsIconSize");
    uiMode = smi->getValue("CoreApps", "UIMode");
    activities = smi->getValue("CoreApps", "KeepActivities");
    listViewIconSize = smi->getValue("CoreApps", "ListViewIconSize");

    // get app's settings
    continuousPage = smi->getValue("CorePDF", "Continuous");
    pageLayout = smi->getValue("CorePDF", "PageLayout");
    zoomMode = smi->getValue("CorePDF", "ZoomMode");
    zoomFactor = smi->getValue("CorePDF", "ZoomFactor");
    windowSize = smi->getValue("CorePDF", "WindowSize");
    windowMaximized = smi->getValue("CorePDF", "WindowMaximized");
}


void corepdf::setupIcons()
{
	ui->open->setIcon(CPrime::ThemeFunc::themeIcon("document-open-symbolic", "quickopen-file", "document-open-symbolic"));
	ui->rotate->setIcon(CPrime::ThemeFunc::themeIcon("object-rotate-right-symbolic", "object-rotate-right", "object-rotate-right"));
	ui->pinIt->setIcon(CPrime::ThemeFunc::themeIcon("bookmark-new-symbolic", "bookmark-new-symbolic", "bookmark-new"));
	ui->shareIt->setIcon(CPrime::ThemeFunc::themeIcon("document-send-symbolic", "document-send-symbolic", "document-send"));
	ui->savePageAsPic->setIcon(CPrime::ThemeFunc::themeIcon("document-save-symbolic", "document-save", "document-save"));
	ui->openmetadataApp->setIcon(CPrime::ThemeFunc::themeIcon("document-properties-symbolic", "document-properties", "document-properties"));
	ui->search->setIcon(CPrime::ThemeFunc::themeIcon("edit-find-symbolic", "edit-find", "edit-find"));
	ui->jumpToPage->setIcon(CPrime::ThemeFunc::themeIcon("view-refresh-symbolic", "view-refresh", "view-refresh"));
	ui->menu->setIcon(CPrime::ThemeFunc::themeIcon("open-menu-symbolic", "application-menu", "open-menu"));
	ui->previousW->setIcon(CPrime::ThemeFunc::themeIcon("go-previous-symbolic", "go-previous", "go-previous"));
	ui->nextW->setIcon(CPrime::ThemeFunc::themeIcon("go-next-symbolic", "go-next", "go-next"));
	ui->closeSearch->setIcon(CPrime::ThemeFunc::themeIcon("window-close-symbolic", "window-close", "window-close"));
}


/**
  * @brief Save application settings
  */
void corepdf::saveSettings()
{
    // save window stats
    smi->setValue("CorePDF", "WindowSize", this->size());
    smi->setValue("CorePDF", "WindowMaximized", this->isMaximized());

	/** Save the if the pages need to be shown continuously */
	smi->setValue("CorePDF", "Continuous", contLytAct->isChecked());

	/** Save the page layout: Single/Facing/Book */
	smi->setValue("CorePDF", "PageLayout", PdfWidget->pageLayout());

	/** Save the zoom mode (Width/Page/Custom) */
	switch (ui->zoomCB->currentIndex())
	{
	case 0:
		smi->setValue("CorePDF", "ZoomMode", QPdfView::FitToWidth);
		break;

	case 1:
		smi->setValue("CorePDF", "ZoomMode", QPdfView::FitInView);
		break;

	default:
		smi->setValue("CorePDF", "ZoomMode", QPdfView::CustomZoom);
		break;
	}

	/** Save the zoom factor */
	smi->setValue("CorePDF", "ZoomFactor", ( int )(PdfWidget->zoomFactor() * 100));
}


/**
  * @brief Loads application shotcuts
  */
void corepdf::shotcuts()
{
	QShortcut *shortcut;

	shortcut = new QShortcut(QKeySequence(Qt::CTRL + Qt::Key_O), this);
	connect(shortcut, &QShortcut::activated, this, &corepdf::openFileDialog);
	shortcut = new QShortcut(QKeySequence(Qt::CTRL + Qt::Key_F), ui->app);
	connect(shortcut, &QShortcut::activated, this, &corepdf::on_search_clicked);
	shortcut = new QShortcut(QKeySequence(Qt::Key_Escape), ui->app);
	connect(shortcut, &QShortcut::activated, this, &corepdf::on_search_clicked);
}


/**
  * @brief Sends file to library
  */
void corepdf::loadFile(const QString& file)
{
	workFilePath = file;
	ui->page->setCurrentIndex(1);

	PdfDoc = new CorePDF::Document(file);
	PdfDoc->load();

	QString errorMsg;
	QString text = "This document is password encrypted. Please enter the password to view it.";
	CorePDF::Document::Error err;

	do
	{
		err = PdfDoc->error();
		switch (err)
		{
		case CorePDF::Document::UnknownError:
			errorMsg = "CorePDF encountered an unknwon error. Aborting...";
			break;

		case CorePDF::Document::FileNotFoundError:
			errorMsg = "CorePDF could not find the file you are trying to open. Did it get deleted?";
			break;

		case CorePDF::Document::IncorrectPasswordError:
		   {
			   QString response = QInputDialog::getText(
				   this,                                                         // Parent
				   "CorePDF | Enter Password",                                   // Dialog title
				   text,                                                         // Dialog text
				   QLineEdit::Password                                           // Echo mode
				   );

			   /* User clicked cancel or entered an empty string */
			   if (response.isEmpty())
			   {
				   QMessageBox::warning(this, "CorePDF | Document Error", "Sorry! CorePDF cannot open this file without a password.");
				   ui->page->setCurrentIndex(0);
				   return;
			   }

			   /* Check if the password is correct */
			   PdfDoc->setPassword(response);

			   /* In case the password is wrong, an alternate text */
			   text = "This password you entered is incorrect. Please try again.";

			   break;
		   }

		/** No error */
		default:
			break;
		}

		if (not errorMsg.isEmpty())
		{
			QMessageBox::critical(
				this,
				"CorePDF | Unrecoverable error",
				errorMsg,
				QMessageBox::StandardButton::Ok
				);

			return;
		}
	} while (err != CorePDF::Document::NoError);

	PdfWidget->setDocument(PdfDoc);

	QString title = PdfDoc->title();

	ui->fileName->setText(title.isEmpty() ? CPrime::FileUtils::baseName(file) : title);

	ui->menu->setEnabled(true);
	ui->rotate->setEnabled(true);

	pageCountLbl->setText(QString("Page 1 of %1").arg(PdfDoc->pageCount()));

	PdfWidget->viewport()->update();
}


void corepdf::closeEvent(QCloseEvent *event)
{
	event->ignore();

	saveSettings();

    if (not workFilePath.isEmpty() && activities)
	{
		CPrime::ActivitiesManage::saveToActivites("corepdf", QStringList() << workFilePath);
	}

	PdfWidget->close();
	event->accept();
}


void corepdf::timerEvent(QTimerEvent *tEvent)
{
	if (delayTimer->timerId() == tEvent->timerId())
	{
		delayTimer->stop();
		PdfWidget->searchText(ui->searchHere->text());
	}

	QWidget::timerEvent(tEvent);
}


void corepdf::loadActivities()
{
	if (not activities)
	{
		ui->activitiesList->setVisible(false);
		return;
	}

	ui->activitiesList->clear();

	QSettings   recentActivity(CPrime::Variables::CC_ActivitiesFilePath(), QSettings::IniFormat);
	QStringList topLevel = recentActivity.childGroups();

	if (topLevel.count())
	{
		topLevel = CPrime::SortFunc::sortDate(topLevel, Qt::DescendingOrder);
	}

	QListWidgetItem *item;

	Q_FOREACH (QString group, topLevel)
	{
		recentActivity.beginGroup(group);
		QStringList keys = recentActivity.childKeys();
		keys = CPrime::SortFunc::sortTime(keys, Qt::DescendingOrder, "hh.mm.ss.zzz");

		Q_FOREACH (QString key, keys)
		{
			QStringList value = recentActivity.value(key).toString().split("\t\t\t");
			if (value[0] == "corepdf")
			{
				QString filePath = value[1];
				if (not CPrime::FileUtils::exists(filePath))
				{
					continue;
				}

				QString baseName = CPrime::FileUtils::baseName(filePath);
				item = new QListWidgetItem(baseName);
				item->setData(Qt::UserRole, filePath);
				item->setToolTip(filePath);
				item->setIcon(CPrime::ThemeFunc::getFileIcon(filePath));
				ui->activitiesList->addItem(item);
			}
		}

		recentActivity.endGroup();
	}
}


void corepdf::on_menu_clicked()
{
	if (ui->sideView->isVisible())
	{
		ui->sideView->setVisible(false);
	}
	else
	{
		ui->sideView->setVisible(true);
	}
}


void corepdf::on_shareIt_clicked()
{
	if (!workFilePath.isNull())
	{
		ShareIT *t = new ShareIT(QStringList() << workFilePath, listViewIconSize, nullptr);
		// Set ShareIT window size
		if (uiMode == 2)
		{
			t->setFixedSize(QGuiApplication::primaryScreen()->size() * .8);
		}
		else
		{
			t->resize(500, 600);
		}

		t->exec();
	}
}


void corepdf::on_pinIt_clicked()
{
	if (!workFilePath.isNull())
	{
		PinIT *pit = new PinIT(QStringList() << workFilePath, this);
		pit->exec();
	}
}


void corepdf::openFileDialog()
{
	QString openFrom = QDir::homePath();

	if (!workFilePath.isNull())
	{
		openFrom = workFilePath;
	}
	QString fn = QFileDialog::getOpenFileName(
		this,
		"CorePDF - Open File",
		openFrom,
		"PDF Files (*.pdf);;All Files (*.*)"
		);

	if (not fn.isEmpty())
	{
		loadFile(fn);
	}
}


void corepdf::on_openmetadataApp_clicked()
{
	CPrime::AppOpenFunc::defaultAppEngine(CPrime::DefaultAppCategory::MetadataViewer, workFilePath, QString());
}


void corepdf::on_savePageAsPic_clicked()
{
	int   pageNo = PdfWidget->pageNavigation()->currentPage();
	QSize pSize  = PdfDoc->pageSize(pageNo).toSize();

	/* Obtain the page zoom factor */
	QInputDialog *zoomDlg = new QInputDialog(this);

	zoomDlg->setInputMode(QInputDialog::IntInput);
	zoomDlg->setLabelText(
		QString(
			"Size of page %1 is %2 x %3 px. Please set the zoom factor below if necesssary."
			).arg(pageNo + 1).arg(pSize.width()).arg(pSize.height())
		);
	zoomDlg->setIntRange(10, 400);
	zoomDlg->setIntStep(10);
	zoomDlg->setIntValue(100);

	Q_FOREACH (QSpinBox *sb, zoomDlg->findChildren<QSpinBox *>())
	{
		if (sb)
		{
			sb->setSuffix(" %");
		}
	}

	if (zoomDlg->exec() == QDialog::Accepted)
	{
		/* Obtain the save file name */
		QStringList fmts;
		Q_FOREACH (QByteArray fmtba, QImageWriter::supportedImageFormats())
		{
			QString fmt = QString::fromLatin1(fmtba);
			fmts << fmt.toUpper() + " Image (*." + fmt + ")";
		}

		QString *str = new QString("PNG Image (*.png)");
		QString fn   = QFileDialog::getSaveFileName(
			nullptr,
			"CorePDF - Save Current Page",
			QDir::homePath() + "/untitled.png",
			fmts.join(";;") + ";;All Files (*.*)",
			str
			);

		if (not fn.isEmpty())
		{
			QString fmt = str->split(" ")[0].toLower();
			if (not fn.endsWith(fmt))
			{
				fn += "." + fmt;
			}

			qDebug() << "Requesting image of size" << pSize * zoomDlg->intValue() << fn;
			PageSaver *saver = new PageSaver(PdfDoc, pageNo, pSize * zoomDlg->intValue() / 100.0, fn);
			saver->start();
		}
	}
}


void corepdf::on_zoomCB_activated(int idx)
{
	if (not PdfDoc or PdfDoc->status() != CorePDF::Document::Ready)
	{
		return;
	}

	if (not PdfWidget)
	{
		return;
	}

	if (idx == cur_zoom_idx)
	{
		return;
	}

	cur_zoom_idx = idx;
	switch (idx)
	{
	case 0:
		PdfWidget->setZoomMode(QPdfView::FitToWidth);
		break;

	case 1:
		PdfWidget->setZoomMode(QPdfView::FitInView);
		break;

	default:
	   {
		   PdfWidget->setZoomMode(QPdfView::CustomZoom);
		   qreal zoom = ui->zoomCB->currentText().replace("%", "").toInt() * 1.0 / 100;
		   PdfWidget->setZoomFactor(zoom);
	   }
	}
}


void corepdf::on_rotate_clicked()
{
	int rotation = ( int )PdfWidget->rotation();

	rotation = (rotation + 1) % 4;

	PdfWidget->setRotation((CorePDF::RenderOptions::Rotation)rotation);
}


void corepdf::updatePageNumber(int pageNo)
{
	pageCountLbl->setText(QString("Page %1 of %2").arg(QString::number(pageNo + 1)).arg(PdfDoc->pageCount()));
}


void corepdf::on_activitiesList_itemDoubleClicked(QListWidgetItem *item)
{
	QString filepath = item->data(Qt::UserRole).toString();

	loadFile(filepath);
}


void corepdf::on_search_clicked()
{
	if (ui->searchB->isVisible())
	{
		ui->searchB->setVisible(false);
		pageCountLbl->setVisible(true);
	}
	else
	{
		ui->searchB->setVisible(true);
		ui->searchHere->setFocus();
		pageCountLbl->setVisible(false);
	}
}


void corepdf::on_closeSearch_clicked()
{
	ui->searchB->setVisible(false);
	pageCountLbl->setVisible(true);
}


void corepdf::jumpPage()
{
	int curPage  = PdfWidget->pageNavigation()->currentPage();
	int maxPages = PdfDoc->pageCount();

	QInputDialog *inDlg = new QInputDialog(this);

	inDlg->setWindowTitle("CorePDF | Goto Page");
	inDlg->setLabelText("Enter the page to which you want to jump:");
	inDlg->setInputMode(QInputDialog::IntInput);
	inDlg->setIntRange(1, maxPages);
	inDlg->setIntValue(curPage + 1);
	inDlg->setIntStep(1);

	QSpinBox *sb = inDlg->findChild<QSpinBox *>();

	if (sb)
	{
		sb->setSuffix(QString(" of %1 pages").arg(maxPages));
	}

	if (inDlg->exec())
	{
		int tgtPg = inDlg->intValue() - 1;
		PdfWidget->pageNavigation()->setCurrentPage(tgtPg);
		pageCountLbl->setText(QString("Page %1 of %2").arg(tgtPg + 1).arg(maxPages));
	}
}


PageSaver::PageSaver(CorePDF::Document *doc, int page, QSize size, const QString &saveName)
	: mDoc(doc)
	, pageNo(page)
	, mSize(size)
	, imgName(saveName)
{
}


void PageSaver::run()
{
	QImage image = mDoc->renderPage(pageNo, mSize, CorePDF::RenderOptions());

	if (not image.isNull())
	{
		QImage page(image.size(), QImage::Format_ARGB32);
		page.fill(Qt::white);

		QPainter p(&page);
		p.drawImage(QRect(QPoint(0, 0), page.size()), image);
		p.end();

		bool ret = page.save(imgName);

		qDebug() << "Saved file" << (ret ? "[SUCCESS]" : "[FAILED]");
		qDebug() << "Image" << image.size();
	}
}
