/**
  * A PDF viewer for C Suite.
  * Copyright 2019 CuboCore Group
  *
  * This file is a part of QtPDF module from Qt-Labs (https://code.qt.io/cgit/qt-labs/qtpdf.git/).
  * Suitable modifications have been made to meet the needs of CorePDF.
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
  * the Free Software Foundation; either version 3 of the License, or
  * (at your option) any later version.
  *
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, vsit http://www.gnu.org/licenses/.
  **/

#include <cprime/filefunc.h>
#include "Document.hpp"

/*
  *
  * Generic class to handle document
  *
  */

CorePDF::Document::Document(const QString &path) : QObject()
{
	mZoom       = 1.0;
	mStatus     = Null;
	mDocError   = NoError;
	mPassNeeded = false;

	mDocPath = QFileInfo(path).absoluteFilePath();

	QFileSystemWatcher *fsw = new QFileSystemWatcher();

	fsw->addPath(mDocPath);
	connect(
		fsw, &QFileSystemWatcher::fileChanged, [ = ]( QString file ) {
		/* File deleted and created again: add it to the watcher */
		if (not fsw->files().contains(file) and QFile::exists(file))
		{
			fsw->addPath(file);
		}

		reload();
	}
		);

	mPdfDoc = nullptr;
}


bool CorePDF::Document::passwordNeeded() const
{
	return mPassNeeded;
}


void CorePDF::Document::setPassword(const QString &password)
{
	if (mPdfDoc->unlock(password.toLatin1(), password.toLatin1()))
	{
		mStatus   = Failed;
		mDocError = IncorrectPasswordError;

		qDebug() << "Invalid password. Please try again.";
		mPassNeeded = true;
		emit statusChanged(Failed);
		emit passwordRequired();

		return;
	}

	mPassNeeded = false;
	mStatus     = Loading;
	mDocError   = NoError;

	emit statusChanged(Loading);

	mPdfDoc->setRenderHint(Poppler::Document::Antialiasing);
	mPdfDoc->setRenderHint(Poppler::Document::TextAntialiasing);
	mPdfDoc->setRenderHint(Poppler::Document::TextHinting);

	for ( int i = 0; i < mPdfDoc->numPages(); i++ )
	{
		Poppler::Page *p    = mPdfDoc->page(i);
		CorePDF::Page *page = new CorePDF::Page(i, p);
		mPages.append(page);

		emit loading(1.0 * i / mPdfDoc->numPages() * 100.0);
	}

	mStatus   = Ready;
	mDocError = NoError;

	emit statusChanged(Ready);
	emit pageCountChanged(mPages.count());
	emit loading(100);
}


QString CorePDF::Document::fileName() const
{
	return CPrime::FileUtils::baseName(mDocPath);
}


QString CorePDF::Document::filePath() const
{
	return CPrime::FileUtils::dirName(mDocPath);
}


QString CorePDF::Document::title() const
{
	return mPdfDoc->title();
}


QString CorePDF::Document::author() const
{
	return mPdfDoc->author();
}


QString CorePDF::Document::creator() const
{
	return mPdfDoc->creator();
}


QString CorePDF::Document::producer() const
{
	return mPdfDoc->producer();
}


QString CorePDF::Document::created() const
{
	return mPdfDoc->creationDate().toString("MMM DD, yyyy hh:mm:ss t AP");
}


int CorePDF::Document::pageCount() const
{
	return mPages.count();
}


QSizeF CorePDF::Document::pageSize(int pageNo) const
{
	if (pageNo >= mPages.count())
	{
		return QSizeF();
	}

	return mPages.at(pageNo)->pageSize() * mZoom;
}


void CorePDF::Document::reload()
{
	mStatus = Null;
	mPages.clear();

	load();

	if (mStatus == Ready)
	{
		qDebug() << "Reload your pages..";
		emit reloadDocument();
	}
}


CorePDF::Document::Status CorePDF::Document::status() const
{
	return mStatus;
}


CorePDF::Document::Error CorePDF::Document::error() const
{
	return mDocError;
}


QImage CorePDF::Document::renderPage(int pageNo, QSize size, const CorePDF::RenderOptions &opts) const
{
	if (pageNo >= mPages.count())
	{
		return QImage();
	}

	return mPages.at(pageNo)->render(size, opts);
}


QImage CorePDF::Document::renderPage(int pageNo, qreal zoomFactor, const CorePDF::RenderOptions &opts) const
{
	if (pageNo >= mPages.count())
	{
		return QImage();
	}

	return mPages.at(pageNo)->render(zoomFactor, opts);
}


CorePDF::Pages CorePDF::Document::pages() const
{
	return mPages;
}


CorePDF::Page * CorePDF::Document::page(int pageNo) const
{
	if (pageNo >= mPages.count())
	{
		return nullptr;
	}

	return mPages.at(pageNo);
}


QImage CorePDF::Document::pageThumbnail(int pageNo) const
{
	if (pageNo >= mPages.count())
	{
		return QImage();
	}

	return mPages.at(pageNo)->thumbnail();
}


QString CorePDF::Document::pageText(int pageNo) const
{
	if (pageNo >= mPages.count())
	{
		return QString();
	}

	return mPages.at(pageNo)->text(QRectF());
}


QString CorePDF::Document::text(int pageNo, QRectF rect) const
{
	if (pageNo >= mPages.count())
	{
		return QString();
	}

	return mPages.at(pageNo)->text(rect);
}


QList<QRectF> CorePDF::Document::search(const QString &query, int pageNo, const RenderOptions &opts) const
{
	if (query.isEmpty())
	{
		return QList<QRectF>();
	}

	if (pageNo >= mPages.count())
	{
		return QList<QRectF>();
	}

	searchRects.clear();

	if (pageNo >= 0 and pageNo < mPages.count())
	{
		for ( QRectF rect: mPages.at(pageNo)->search(query, opts))
		{
			searchRects << QRect(rect.x() * mZoom, rect.y() * mZoom, rect.width() * mZoom, rect.height() * mZoom);
		}
	}

	return searchRects;
}


qreal CorePDF::Document::zoomForWidth(int pageNo, qreal width) const
{
	if (pageNo >= mPages.count())
	{
		return 1.0;
	}

	return 1.0 * width / mPages.at(pageNo)->pageSize().width();
}


qreal CorePDF::Document::zoomForHeight(int pageNo, qreal height) const
{
	if (pageNo >= mPages.count())
	{
		return 1.0;
	}

	return 1.0 * height / mPages.at(pageNo)->pageSize().height();
}


void CorePDF::Document::setZoom(qreal zoom)
{
	mZoom = zoom;
}


void CorePDF::Document::load()
{
	mStatus = Loading;
	emit statusChanged(Loading);

	if (not QFile::exists(mDocPath))
	{
		mStatus   = Failed;
		mDocError = FileNotFoundError;
		emit statusChanged(Failed);

		return;
	}

	mPdfDoc = Poppler::Document::load(mDocPath);

	if (not mPdfDoc)
	{
		mStatus   = Failed;
		mDocError = UnknownError;
		qDebug() << "Poppler::Document load failed";
		emit statusChanged(Failed);

		return;
	}

	if (mPdfDoc->isLocked())
	{
		mStatus   = Failed;
		mDocError = IncorrectPasswordError;
		qDebug() << "Poppler::Document is locked";
		mPassNeeded = true;
		emit passwordRequired();
		emit statusChanged(Failed);
		return;
	}

	mPdfDoc->setRenderHint(Poppler::Document::Antialiasing);
	mPdfDoc->setRenderHint(Poppler::Document::TextAntialiasing);
	mPdfDoc->setRenderHint(Poppler::Document::TextHinting);

	for ( int i = 0; i < mPdfDoc->numPages(); i++ )
	{
		Poppler::Page *p    = mPdfDoc->page(i);
		CorePDF::Page *page = new CorePDF::Page(i, p);
		mPages.append(page);

		emit loading(1.0 * i / mPdfDoc->numPages() * 100.0);
	}

	mStatus   = Ready;
	mDocError = NoError;

	emit statusChanged(Ready);
	emit pageCountChanged(mPages.count());
	emit loading(100);
}


void CorePDF::Document::close()
{
	mStatus = Unloading;
	mPages.clear();
	mZoom = 1.0;

	delete mPdfDoc;
}


/*
  *
  * Generic class to handle document page
  *
  */

CorePDF::Page::Page(int pgNo, Poppler::Page *p)
{
	mPageNo = pgNo;
	m_page  = p;
}


CorePDF::Page::~Page()
{
	delete m_page;
}


int CorePDF::Page::pageNo()
{
	return mPageNo;
}


QImage CorePDF::Page::render(QSize pSize, const CorePDF::RenderOptions &opts) const
{
	qreal wZoom = 1.0 * pSize.width() / m_page->pageSizeF().width();
	qreal hZoom = 1.0 * pSize.height() / m_page->pageSizeF().height();

	return m_page->renderToImage(72 * wZoom, 72 * hZoom, -1, -1, -1, -1, ( Poppler::Page::Rotation )opts.rotation());
}


QImage CorePDF::Page::render(qreal zoomFactor, const RenderOptions &opts) const
{
	return m_page->renderToImage(72 * zoomFactor, 72 * zoomFactor, -1, -1, -1, -1, ( Poppler::Page::Rotation )opts.rotation());
}


QString CorePDF::Page::pageText() const
{
	return text(QRectF());
}


QString CorePDF::Page::text(QRectF rect) const
{
	return m_page->text(rect);
}


QList<QRectF> CorePDF::Page::search(const QString &query, const CorePDF::RenderOptions &opts) const
{
	return m_page->search(
		query,                                                                                      // Search text
		Poppler::Page::IgnoreCase | Poppler::Page::IgnoreDiacritics,                                // Case insensitive
		( Poppler::Page::Rotation )opts.rotation()                                                  // Rotation
		);
}


QSizeF CorePDF::Page::pageSize(qreal zoom) const
{
	return m_page->pageSizeF() * zoom;
}


QImage CorePDF::Page::thumbnail() const
{
	return m_page->thumbnail();
}
