/**
  * A PDF viewer for C Suite.
  * Copyright 2019 CuboCore Group
  *
  * This file is a part of QtPDF module from Qt-Labs (https://code.qt.io/cgit/qt-labs/qtpdf.git/).
  * Suitable modifications have been made to meet the needs of CorePDF.
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
  * the Free Software Foundation; either version 3 of the License, or
  * (at your option) any later version.
  *
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, vsit http://www.gnu.org/licenses/.
  **/

#pragma once

#include <QtCore>
#include <QtGui>
#include <QtWidgets>

#include "RenderOptions.hpp"

#include <poppler/qt5/poppler-qt5.h>

namespace CorePDF {
class Page;
typedef QList<CorePDF::Page *> Pages;

class Document : public QObject {
	Q_OBJECT;

public:
	enum Status
	{
		Null,
		Loading,
		Ready,
		Unloading,
		Failed
	};
	Q_ENUM(CorePDF::Document::Status)

	enum Error
	{
		NoError,
		UnknownError,
		FileNotFoundError,
		IncorrectPasswordError
	};
	Q_ENUM(CorePDF::Document::Error)

	enum MetaDataField
	{
		Title,
		Subject,
		Author,
		Keywords,
		Producer,
		Creator,
		CreationDate,
		ModificationDate
	};
	Q_ENUM(MetaDataField)

	explicit Document(const QString &docPath);

	/* Check if a password is needed */
	bool passwordNeeded() const;

	/* Set a password */
	void setPassword(const QString &password);

	/* Document File Name and File Path */
	QString fileName() const;
	QString filePath() const;

	/* Pdf Info / Metadata */
	QString title() const;
	QString author() const;
	QString creator() const;
	QString producer() const;
	QString created() const;

	/* Number of pages */
	int pageCount() const;

	/* Size of the page */
	QSizeF pageSize(int pageNo) const;

	/* Reload the current document */
	void reload();

	/* PDF load status */
	CorePDF::Document::Status status() const;
	CorePDF::Document::Error error() const;

	/* Render and return a page */
	QImage renderPage(int, QSize, const RenderOptions &) const;
	QImage renderPage(int, qreal zoomFactor, const RenderOptions &) const;

	/* Page Related */
	Pages pages() const;
	Page * page(int pageNo) const;
	QImage pageThumbnail(int pageNo) const;

	/* Page Text */
	QString pageText(int pageNo) const;

	/* Text of a Selection rectangle */
	QString text(int pageNo, QRectF) const;

	/* Search for @query in @pageNo or all pages */
	QList<QRectF> search(const QString &query, int pageNo, const RenderOptions &opts) const;

	qreal zoomForWidth(int pageNo, qreal width) const;
	qreal zoomForHeight(int pageNo, qreal width) const;

	void setZoom(qreal zoom);

public Q_SLOTS:
	void load();
	void close();

private:
	/* Pointer to our actual poppler document */
	Poppler::Document *mPdfDoc;

	QString mDocPath;
	Pages mPages;

	mutable QList<QRectF> searchRects;

	qreal mZoom;

	CorePDF::Document::Status mStatus;
	CorePDF::Document::Error mDocError;
	bool mPassNeeded;

Q_SIGNALS:
	void passwordRequired();
	void reloadDocument();
	void statusChanged(Document::Status status);
	void pageCountChanged(int pageCount);

	void loading(int);
};

class Page {
public:
	Page(int, Poppler::Page *);
	~Page();

	/* Page number of this page */
	int pageNo();

	/* Render and return a page */
	QImage render(QSize, const RenderOptions &) const;
	QImage render(qreal zoomFactor, const RenderOptions &) const;

	/* Page Text */
	QString pageText() const;

	/* Text of a Selection rectangle */
	QString text(QRectF) const;

	/* Search for @query in @pageNo or all pages */
	QList<QRectF> search(const QString &query, const RenderOptions &opts) const;

	/* Size of the page */
	QSizeF pageSize(qreal zoom = 1.0) const;

	/* Thumbnail of the page */
	QImage thumbnail() const;

private:
	int mPageNo = -1;
	Poppler::Page *m_page;
};
}
