/**
  * A PDF viewer for C Suite.
  * Copyright 2019 CuboCore Group
  *
  * This file is a part of QtPDF module from Qt-Labs (https://code.qt.io/cgit/qt-labs/qtpdf.git/).
  * Suitable modifications have been made to meet the needs of CorePDF.
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
  * the Free Software Foundation; either version 3 of the License, or
  * (at your option) any later version.
  *
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, vsit http://www.gnu.org/licenses/.
  **/

#include "qpdfview.h"
#include "qpdfview_p.h"

#include "Renderer.hpp"
#include "Document.hpp"
#include "Navigation.hpp"

#include <QGuiApplication>
#include <QScreen>
#include <QScrollBar>
#include <QScroller>
#include <QMessageBox>

QT_BEGIN_NAMESPACE

QPdfView::QPdfView(QWidget *parent)
	: QAbstractScrollArea(parent)
	, impl(new QPdfViewImpl(this))
{
	impl->init();

	/** Piggy-back the matchesFound, and searchComplete signals from the search thread */
	connect(impl->m_searchThread, &PdfSearchThread::matchesFound, this, &QPdfView::matchesFound);
	connect(impl->m_searchThread, &PdfSearchThread::searchComplete, this, &QPdfView::searchComplete);

	connect(impl->m_searchThread, &PdfSearchThread::resultsReady, [this]( int page ) {
		impl->searchRects[page] = impl->m_searchThread->results(page);
		viewport()->update();
	});

	connect(impl->m_searchThread, &PdfSearchThread::resultsReady, [this]( int page ) {
		impl->searchRects[page] = impl->m_searchThread->results(page);
		viewport()->update();
	});

	connect(impl->m_pageNavigation, &CorePDF::Navigation::currentPageChanged, this, [this]( int page ){
		impl->currentPageChanged(page);
		impl->m_searchThread->searchPage(page);
	});

	/* Setup Page Renderer */
	connect(impl->m_pageRenderer, &CorePDF::Renderer::pageRendered, [ = ]( int ) {
		viewport()->update();
	});

	verticalScrollBar()->setSingleStep(20);
	horizontalScrollBar()->setSingleStep(20);

	QScroller::grabGesture(this);

	setFrameStyle(QFrame::NoFrame);

	impl->calculateViewport();
}


QPdfView::~QPdfView()
{
}


void QPdfView::setDocument(CorePDF::Document *document)
{
	if (impl->m_document == document)
	{
		return;
	}

	if (impl->m_document)
	{
		disconnect(impl->m_documentStatusChangedConnection);
	}

	impl->m_document = document;
	emit documentChanged(impl->m_document);

	if (impl->m_document)
	{
		impl->m_documentStatusChangedConnection = connect(impl->m_document, &CorePDF::Document::statusChanged, this, [this](){
			impl->documentStatusChanged();
		});
	}

	impl->m_pageNavigation->setDocument(impl->m_document);
	impl->m_pageRenderer->setDocument(impl->m_document);
	impl->m_searchThread->setDocument(document);

	impl->documentStatusChanged();
}


CorePDF::Document *QPdfView::document() const
{
	return impl->m_document;
}


CorePDF::Navigation *QPdfView::pageNavigation() const
{
	return impl->m_pageNavigation;
}


bool QPdfView::isLayoutContinuous() const
{
	return impl->m_continuous;
}


void QPdfView::setLayoutContinuous(bool yes)
{
	if (impl->m_continuous == yes)
	{
		return;
	}

	impl->m_continuous = yes;
	impl->invalidateDocumentLayout();

	emit layoutContinuityChanged(yes);
}


QPdfView::PageLayout QPdfView::pageLayout() const
{
	return impl->m_pageLayout;
}


void QPdfView::setPageLayout(PageLayout layout)
{
	if (impl->m_pageLayout == layout)
	{
		return;
	}

	impl->m_pageLayout = layout;
	impl->invalidateDocumentLayout();

	emit pageLayoutChanged(impl->m_pageLayout);
}


QPdfView::ZoomMode QPdfView::zoomMode() const
{
	return impl->m_zoomMode;
}


void QPdfView::setZoomMode(ZoomMode mode)
{
	if (impl->m_zoomMode == mode)
	{
		return;
	}

	impl->m_zoomMode = mode;
	impl->invalidateDocumentLayout();

	emit zoomModeChanged(impl->m_zoomMode);
}


qreal QPdfView::zoomFactor() const
{
	return impl->m_zoomFactor;
}


void QPdfView::setZoomFactor(qreal factor)
{
	if (impl->m_zoomFactor == factor)
	{
		return;
	}

	impl->m_zoomFactor = factor;

	if (impl->m_zoomMode == QPdfView::CustomZoom)
	{
		impl->invalidateDocumentLayout();
		emit zoomFactorChanged(impl->m_zoomFactor);
	}
}


int QPdfView::pageSpacing() const
{
	return impl->m_pageSpacing;
}


void QPdfView::setPageSpacing(int spacing)
{
	if (impl->m_pageSpacing == spacing)
	{
		return;
	}

	impl->m_pageSpacing = spacing;
	impl->invalidateDocumentLayout();

	emit pageSpacingChanged(impl->m_pageSpacing);
}


QMargins QPdfView::documentMargins() const
{
	return impl->m_documentMargins;
}


void QPdfView::setDocumentMargins(QMargins margins)
{
	if (impl->m_documentMargins == margins)
	{
		return;
	}

	impl->m_documentMargins = margins;
	impl->invalidateDocumentLayout();

	emit documentMarginsChanged(impl->m_documentMargins);
}


void QPdfView::searchText(const QString &str)
{
	impl->searchRects.clear();

	impl->m_searchThread->setSearchString(str);
	impl->m_searchThread->searchPage(impl->m_pageNavigation->currentPage());
}


CorePDF::RenderOptions::Rotation QPdfView::rotation() const
{
	return impl->m_renderOpts.rotation();
}


void QPdfView::setRotation(CorePDF::RenderOptions::Rotation rot)
{
	impl->m_renderOpts.setRotation(rot);
	impl->invalidateDocumentLayout();

	emit rotationChanged(rot);
}


void QPdfView::paintEvent(QPaintEvent *event)
{
	if (!impl->m_document) {
		QAbstractScrollArea::paintEvent(event);
		return;
	}

	QPainter *painter = new QPainter(viewport());

	painter->fillRect(event->rect(), palette().brush(QPalette::Dark));
	painter->translate(-impl->m_viewport.x(), -impl->m_viewport.y());

	for ( int i = 0; i < impl->m_document->pageCount(); i++ )
	{
		if (not impl->m_documentLayout.pageGeometries.contains(i))
		{
			continue;
		}

		const QRect pgRect = impl->m_documentLayout.pageGeometries[i];
		if (pgRect.isNull())
		{
			continue;
		}

		if (pgRect.intersects(impl->m_viewport))
		{
			painter->fillRect(pgRect, Qt::white);
			QImage img = impl->m_pageRenderer->requestPage(i, pgRect.size(), impl->m_renderOpts);

			if (img.width() and img.height())
			{
				impl->paintSearchRects(i, img);
				painter->drawImage(pgRect.topLeft(), img);
			}
		}
	}

	painter->end();
}


void QPdfView::resizeEvent(QResizeEvent *event)
{
	QAbstractScrollArea::resizeEvent(event);

	impl->updateScrollBars();
	impl->calculateViewport();
}


void QPdfView::scrollContentsBy(int dx, int dy)
{
	QAbstractScrollArea::scrollContentsBy(dx, dy);

	impl->calculateViewport();
}


void QPdfView::keyPressEvent(QKeyEvent *kEvent)
{
	switch (kEvent->key())
	{
	case Qt::Key_Right:
		/* Go to next page */
		impl->m_pageNavigation->setCurrentPage(impl->m_pageNavigation->currentPage() + 1);
		break;

	case Qt::Key_Left:
		/* Go to previous page */
		impl->m_pageNavigation->setCurrentPage(impl->m_pageNavigation->currentPage() - 1);
		break;

	case Qt::Key_Down:
		/** Single row layout. */
		if (not impl->m_continuous)
		{
			/** Scroll down past the end */
			if (verticalScrollBar()->value() == verticalScrollBar()->maximum())
			{
				/** Already at the bottom of last page */
				if (impl->m_pageNavigation->currentPage() == impl->m_document->pageCount() - 1)
				{
					break;
				}

				impl->m_pageNavigation->setCurrentPage(impl->m_pageNavigation->currentPage() + 1);
				break;
			}
		}

		QAbstractScrollArea::keyPressEvent(kEvent);
		break;

	case Qt::Key_Up:
		/** Single row layout. */
		if (not impl->m_continuous)
		{
			/** Scroll down past the end */
			if (verticalScrollBar()->value() == verticalScrollBar()->minimum())
			{
				/** Already at the top of first page */
				if (impl->m_pageNavigation->currentPage() == 0)
				{
					break;
				}

				impl->m_pageNavigation->setCurrentPage(impl->m_pageNavigation->currentPage() - 1);
				verticalScrollBar()->setValue(verticalScrollBar()->maximum());
				break;
			}
		}

		QAbstractScrollArea::keyPressEvent(kEvent);
		break;

	case Qt::Key_Space:
		/* Move by approximately 90% of the viewport height */
		verticalScrollBar()->setValue(verticalScrollBar()->value() + viewport()->height() * 0.9);
		break;

	case Qt::Key_Home:
		/* Go to first page */
		verticalScrollBar()->setValue(0);
		break;

	case Qt::Key_End:
		/* Go to end of the document */
		verticalScrollBar()->setValue(verticalScrollBar()->maximum());
		break;

	case Qt::Key_Plus:
		/* Zoom In */
		setZoomFactor(impl->m_zoomFactor * 1.10);
		break;

	case Qt::Key_Minus:
		/* Zoom Out */
		setZoomFactor(impl->m_zoomFactor / 1.10);
		break;

	default:
		/* Default action */
		QAbstractScrollArea::keyPressEvent(kEvent);
		break;
	}
}


void QPdfView::wheelEvent(QWheelEvent *wEvent)
{
	if (wEvent->buttons() & Qt::RightButton || (wEvent->modifiers() & Qt::ControlModifier))
	{
		QPoint numDegrees = wEvent->angleDelta() / 8;

		int steps = numDegrees.y() / 15;

		QList<qreal> zooms = { 0.1, 0.2, 0.5, 0.75, 1.0, 1.5, 2.0, 4.0 };

		/** - value => scroll up -> increase zoom */
		/** - value => scroll down -> decrease zoom */

		if (impl->m_zoomMode == QPdfView::FitToWidth)
		{
			setZoomMode(QPdfView::CustomZoom);
			setZoomFactor(1.0);
		}
		else if (impl->m_zoomMode == QPdfView::FitInView)
		{
			setZoomMode(QPdfView::CustomZoom);
			setZoomFactor(0.5);
		}
		else
		{
			setZoomMode(QPdfView::CustomZoom);
		}

		if (steps < 0)
		{
			if (impl->m_zoomFactor >= 0.2)
			{
				setZoomFactor(zooms.at(zooms.indexOf(impl->m_zoomFactor) - 1));
			}
		}
		else
		{
			if (impl->m_zoomFactor < 4.0)
			{
				setZoomFactor(zooms.at(zooms.indexOf(impl->m_zoomFactor) + 1));
			}
		}

		return;
	}

	/** Single row layout. */
	if (not impl->m_continuous)
	{
		QPoint numDegrees = wEvent->angleDelta() / 8;
		int    steps      = numDegrees.y() / 15;

		/** Scroll down past the end */
		if ((verticalScrollBar()->value() == verticalScrollBar()->maximum()) and (steps < 0))
		{
			impl->m_pageNavigation->setCurrentPage(impl->m_pageNavigation->currentPage() + 1);
			return;
		}

		/** Scroll up before the beginning */
		if ((verticalScrollBar()->value() == verticalScrollBar()->minimum()) and (steps > 0))
		{
			impl->m_pageNavigation->setCurrentPage(impl->m_pageNavigation->currentPage() - 1);
			return;
		}
	}

	QAbstractScrollArea::wheelEvent(wEvent);
}


QT_END_NAMESPACE
