/**
  * A PDF viewer for C Suite.
  * Copyright 2019 CuboCore Group
  *
  * This file is a part of QtPDF module from Qt-Labs (https://code.qt.io/cgit/qt-labs/qtpdf.git/).
  * Suitable modifications have been made to meet the needs of CorePDF.
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
  * the Free Software Foundation; either version 3 of the License, or
  * (at your option) any later version.
  *
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, vsit http://www.gnu.org/licenses/.
  **/

#include "qpdfview_p.h"

#include "Renderer.hpp"
#include "Document.hpp"
#include "Navigation.hpp"

QT_BEGIN_NAMESPACE

PdfSearchThread::PdfSearchThread(QObject *parent)
	: QThread(parent)
	, mDoc(nullptr)
	, stop(false)
	, stop_others(false)
	, matchCount(0)
{
	connect(this, &PdfSearchThread::pendingRestart, [this]() {
		stop_others = false;
		start();
	});
}


PdfSearchThread::~PdfSearchThread()
{
	stop = true;
	wait();

	matchCount = 0;
	emit matchesFound(matchCount);

	mDoc->deleteLater();
}


void PdfSearchThread::setDocument(CorePDF::Document *doc)
{
	if (mDoc == doc)
	{
		return;
	}

	stop = true;

	matchCount = 0;
	emit matchesFound(matchCount);

	needle = QString();
	pages.clear();
	mResults.clear();

	mDoc = doc;
}


void PdfSearchThread::setSearchString(const QString &str)
{
	if (needle == str.toLower())
	{
		return;
	}

	stop = true;

	matchCount = 0;
	emit matchesFound(matchCount);

	needle = str.toLower();
	pages.clear();
	mResults.clear();
}


void PdfSearchThread::searchPage(int pageNo)
{
	/** Already queued */
	if (pages.contains(pageNo))
	{
		return;
	}

	/** Already searched */
	if (mResults.contains(pageNo))
	{
		return;
	}

	/** Empty search string */
	if (needle.isEmpty())
	{
		return;
	}

	/** Add to the list */
	pages << pageNo;

	stop = false;

	if (isRunning())
	{
		stop_others = true;
	}

	if (not isRunning())
	{
		stop        = false;
		stop_others = false;
		start();
	}
}


QVector<QRectF> PdfSearchThread::results(int pageNo)
{
	return mResults.value(pageNo);
}


void PdfSearchThread::run()
{
	/** Pages requested by the user */
	while (pages.count())
	{
		if (stop)
		{
			return;
		}

		int curPage = pages.takeFirst();

		QList<QRectF> _results = mDoc->search(needle, curPage, CorePDF::RenderOptions());
		matchCount += _results.count();

		emit matchesFound(matchCount);
		mResults[curPage] = QVector<QRectF>::fromList(_results);

		/** If we've stopped, then we don't want any signals being emitted */
		if (not stop)
		{
			emit resultsReady(curPage);
		}
	}

	/** Search all pages from beginning */
	for (int pg = 0; pg < mDoc->pageCount(); pg++)
	{
		if (stop)
		{
			return;
		}

		if (stop_others)
		{
			emit pendingRestart();
			return;
		}

		/** Already finished with @pg */
		if (mResults.contains(pg))
		{
			continue;
		}

		QList<QRectF> _results = mDoc->search(needle, pg, CorePDF::RenderOptions());

		matchCount += _results.count();
		emit matchesFound(matchCount);
		mResults[pg] = QVector<QRectF>::fromList(_results);

		/** If we've stopped, then we don't want any signals being emitted */
		if (not stop)
		{
			emit resultsReady(pg);
		}
	}

	/** We've reached here ==> Search of all pages must be complete. */
	if (mResults.keys().count() == mDoc->pageCount())
	{
		emit searchComplete(matchCount);
	}
}


QPdfViewImpl::QPdfViewImpl(QPdfView *view)
	: m_document(nullptr)
	, m_pageNavigation(nullptr)
	, m_pageRenderer(nullptr)
	, m_continuous(true)
	, m_pageLayout(QPdfView::SinglePage)
	// , m_pageLayout( QPdfView::BookView )
	, m_zoomMode(QPdfView::CustomZoom)
	, m_zoomFactor(1.0)
	, m_pageSpacing(3)
	, m_documentMargins(6, 6, 6, 6)
	, m_blockPageScrolling(false)
	, m_screenResolution(QGuiApplication::primaryScreen()->logicalDotsPerInch() / 72.0)
	, view(view)
{
	m_renderOpts.setRotation(CorePDF::RenderOptions::Rotate0);
	m_renderOpts.setRenderFlags(CorePDF::RenderOptions::RenderAnnotations);
}

QPdfViewImpl::~QPdfViewImpl()
{
	delete m_searchThread;
	delete m_pageNavigation;
	delete m_pageRenderer;
}


void QPdfViewImpl::init()
{
	m_pageNavigation = new CorePDF::Navigation(view);
	m_pageRenderer   = new CorePDF::Renderer(view);
	m_searchThread   = new PdfSearchThread(view);
}


void QPdfViewImpl::documentStatusChanged()
{
	updateDocumentLayout();
	view->viewport()->update();
}


void QPdfViewImpl::currentPageChanged(int currentPage)
{
	if (m_blockPageScrolling)
	{
		return;
	}

	view->verticalScrollBar()->setValue(yPositionForPage(currentPage));

	if (m_pageLayout != QPdfView::SinglePage)
	{
		if (m_documentLayout.pageGeometries.contains(currentPage))
		{
			view->horizontalScrollBar()->setValue(m_documentLayout.pageGeometries[currentPage].x());
		}
	}

	/** Single row display */
	if (not m_continuous)
	{
		switch (m_pageLayout)
		{
		case QPdfView::SinglePage:

			invalidateDocumentLayout();
			break;

		case QPdfView::FacingPages:

			/** if currentPage is even, then we will go to the next row */
			if (not m_documentLayout.pageGeometries.contains(currentPage))
			{
				invalidateDocumentLayout();
			}

			break;

		case QPdfView::BookView:

			/** if currentPage is odd or z, then we will go to the next row */
			if (not m_documentLayout.pageGeometries.contains(currentPage))
			{
				invalidateDocumentLayout();
			}

			break;

		case QPdfView::OverView:

			/** Quite complicated - To be done later? */
			/** Need to know how mange pages per row we are rendering */

			break;
		}
	}
}


void QPdfViewImpl::calculateViewport()
{
	const int x      = view->horizontalScrollBar()->value();
	const int y      = view->verticalScrollBar()->value();
	const int width  = view->viewport()->width();
	const int height = view->viewport()->height();

	setViewport(QRect(x, y, width, height));
}


void QPdfViewImpl::setViewport(QRect viewport)
{
	if (m_viewport == viewport)
	{
		return;
	}

	const QSize oldSize = m_viewport.size();

	m_viewport = viewport;

	if (oldSize != m_viewport.size())
	{
		updateDocumentLayout();

		if (m_zoomMode != QPdfView::CustomZoom)
		{
			view->viewport()->update();
		}
	}

	if (m_continuous)
	{
		// An imaginary, 2px height line at the upper half of the viewport, which is used to
		// determine which page is currently located there -> we propagate that as 'current' page
		// to the CorePDF::Navigation object
		const QRect currentPageLine(m_viewport.x(), m_viewport.y() + m_viewport.height() * 0.4, m_viewport.width(), 2);

		int currentPage = 0;
		for (auto it = m_documentLayout.pageGeometries.cbegin(); it != m_documentLayout.pageGeometries.cend(); ++it)
		{
			const QRect pageGeometry = it.value();
			if (pageGeometry.intersects(currentPageLine))
			{
				currentPage = it.key();
				break;
			}
		}

		if (currentPage != m_pageNavigation->currentPage())
		{
			m_blockPageScrolling = true;
			m_pageNavigation->setCurrentPage(currentPage);
			m_blockPageScrolling = false;
		}
	}
}


void QPdfViewImpl::updateScrollBars()
{
	const QSize p = view->viewport()->size();
	const QSize v = m_documentLayout.documentSize;

	view->horizontalScrollBar()->setRange(0, v.width() - p.width());
	view->horizontalScrollBar()->setPageStep(p.width());
	view->verticalScrollBar()->setRange(0, v.height() - p.height());
	view->verticalScrollBar()->setPageStep(p.height());
}


void QPdfViewImpl::invalidateDocumentLayout()
{
	updateDocumentLayout();
	view->viewport()->update();
}


QPdfViewImpl::DocumentLayout QPdfViewImpl::calculateDocumentLayout() const
{
	switch (m_pageLayout)
	{
	/** */
	case QPdfView::SinglePage:
		return calculateDocumentLayoutSingle();

	/** */
	case QPdfView::FacingPages:
		return calculateDocumentLayoutFacing();

	/** */
	case QPdfView::BookView:
		return calculateDocumentLayoutBook();

	/** */
	case QPdfView::OverView:
		return calculateDocumentLayoutOverview();
	}

	return QPdfViewImpl::DocumentLayout();
}


QPdfViewImpl::DocumentLayout QPdfViewImpl::calculateDocumentLayoutSingle() const
{
	DocumentLayout documentLayout;

	if (!m_document || (m_document->status() != CorePDF::Document::Ready))
	{
		return documentLayout;
	}

	QHash<int, QRect> pageGeometries;

	const int pageCount = m_document->pageCount() - 1;

	int totalWidth = 0;

	const int startPage = (m_continuous ? 0         : m_pageNavigation->currentPage());
	const int endPage   = (m_continuous ? pageCount : m_pageNavigation->currentPage());

	const int horizMargins = m_documentMargins.left() + m_documentMargins.right();
	int       pageY        = m_documentMargins.top();

	// calculate page sizes
	for (int page = startPage; page <= endPage; ++page)
	{
		QSizeF pageSize = m_document->pageSize(page) * m_screenResolution;

		switch (m_renderOpts.rotation())
		{
		/* 90 degree rotated */
		case CorePDF::RenderOptions::Rotate90:
		case CorePDF::RenderOptions::Rotate270:
			/** Swap width <-> height */
			pageSize.transpose();

		default:
			break;
		}

		if (m_zoomMode == QPdfView::CustomZoom)
		{
			pageSize *= m_zoomFactor;
		}

		else if (m_zoomMode == QPdfView::FitToWidth)
		{
			const qreal factor = (qreal(m_viewport.width() - horizMargins) / qreal(pageSize.width()));
			pageSize *= factor;
		}

		else if (m_zoomMode == QPdfView::FitInView)
		{
			const int   margins = (m_continuous ? 0 : m_documentMargins.bottom() + m_documentMargins.top() - m_pageSpacing);
			const QSize viewportSize(m_viewport.size() - QSize(horizMargins, m_pageSpacing + margins));

			pageSize += QSizeF(0, m_pageSpacing);
			pageSize  = pageSize.scaled(viewportSize, Qt::KeepAspectRatio);
		}

		totalWidth           = qMax(totalWidth, pageSize.toSize().width() + horizMargins);
		pageGeometries[page] = QRect(QPoint(0, 0), pageSize.toSize());
	}

	// calculate page positions
	for (int page = startPage; page <= endPage; ++page)
	{
		const QSize pageSize = pageGeometries[page].size();

		// center horizontal inside the viewport
		const int pageX = (qMax(totalWidth, m_viewport.width()) - pageSize.width()) / 2;

		pageGeometries[page].moveTopLeft(QPoint(pageX, pageY));

		pageY += pageSize.height() + m_pageSpacing;
	}

	/** We added an amount 'm_pageSpacing' amount extra */
	pageY += m_documentMargins.bottom() - m_pageSpacing;

	documentLayout.pageGeometries = pageGeometries;

	// calculate overall document size
	documentLayout.documentSize = QSize(totalWidth, pageY);

	return documentLayout;
}


QPdfViewImpl::DocumentLayout QPdfViewImpl::calculateDocumentLayoutFacing() const
{
	DocumentLayout documentLayout;

	if (!m_document || (m_document->status() != CorePDF::Document::Ready))
	{
		return documentLayout;
	}

	QHash<int, QRect> pageGeometries;

	const int pageCount = m_document->pageCount() - 1;
	const int curPage   = m_pageNavigation->currentPage();

	int totalWidth = 0;

	int startPage = 0;
	int endPage   = pageCount;

	if (not m_continuous)
	{
		if (curPage % 2 == 0)
		{
			startPage = curPage;
			endPage   = curPage;
		}

		else
		{
			startPage = curPage - 1;
			endPage   = curPage - 1;
		}
	}

	const int horizMargins = m_documentMargins.left() + m_documentMargins.right();

	// calculate page sizes
	for (int page = startPage; page <= endPage; page += 2)
	{
		QSizeF pageSize;

		QSizeF p1 = m_document->pageSize(page) * m_screenResolution;
		QSizeF p2 = m_document->pageSize(page + 1) * m_screenResolution;

		switch (m_renderOpts.rotation())
		{
		/* 90 degree rotated */
		case CorePDF::RenderOptions::Rotate90:
		case CorePDF::RenderOptions::Rotate270:
			/** Swap width <-> height */
			p1.transpose();
			p2.transpose();

		default:
			break;
		}

		/** Page size is sum of individual pages, suitable pre-rotated */
		pageSize = QSizeF(p1.width() + p2.width(), qMax(p1.height(), p2.height()));

		if (m_zoomMode == QPdfView::CustomZoom)
		{
			pageSize *= m_zoomFactor;
			p1       *= m_zoomFactor;
			p2       *= m_zoomFactor;
			pageSize += QSizeF(m_pageSpacing, 0);
		}

		else if (m_zoomMode == QPdfView::FitToWidth)
		{
			const qreal factor = qreal(m_viewport.width() - horizMargins - m_pageSpacing) / qreal(pageSize.width());
			pageSize *= factor;
			p1       *= (factor / (p2.isValid() ? 1.0 : 2.0));
			p2       *= factor;
			if (p2.isValid())
			{
				pageSize += QSizeF(m_pageSpacing, 0);
			}
		}

		else if (m_zoomMode == QPdfView::FitInView)
		{
			const int   margins = (m_continuous ? 0 : m_documentMargins.bottom() + m_documentMargins.top() - m_pageSpacing);
			const QSize viewportSize(m_viewport.size() - QSize(horizMargins, m_pageSpacing + margins));

			pageSize = pageSize.scaled(viewportSize, Qt::KeepAspectRatio) + QSizeF(m_pageSpacing, 0);
			if (not p2.isValid())
			{
				p1 = QSizeF(pageSize.width(), pageSize.height());
			}

			else
			{
				p1        = QSizeF(pageSize.width() / 2.0, pageSize.height());
				p2        = QSizeF(pageSize.width() / 2.0, pageSize.height());
				pageSize += QSizeF(m_pageSpacing, 0);
			}
		}

		totalWidth = qMax(totalWidth, pageSize.toSize().width() + horizMargins);

		pageGeometries[page] = QRect(QPoint(0, 0), p1.toSize());

		if (not not p2.isValid())
		{
			pageGeometries[page + 1] = QRect(QPoint(0, 0), p2.toSize());
		}
	}

	int pageY = m_documentMargins.top();

	// calculate page positions
	for (int page = startPage; page <= endPage; page += 2)
	{
		const QSize size1 = pageGeometries.value(page).size();
		const QSize size2 = pageGeometries.value(page + 1).size();

		const int pageHeight = qMax(size1.height(), size2.height());

		/** center horizontal inside the viewport */
		const int pageX = (qMax(totalWidth, m_viewport.width()) - size1.width() - (size2.isNull() ? 0 : size2.width() + m_pageSpacing)) / 2;
		pageGeometries[page].moveTopLeft(QPoint(pageX, pageY));

		if (size2.isValid())
		{
			pageGeometries[page + 1].moveTopLeft(QPoint(pageX + size1.width() + m_pageSpacing, pageY));
		}

		pageY += pageHeight + m_pageSpacing;
	}

	pageY += m_documentMargins.bottom() - m_pageSpacing;

	documentLayout.pageGeometries = pageGeometries;

	// calculate overall document size
	documentLayout.documentSize = QSize(totalWidth, pageY);

	return documentLayout;
}


QPdfViewImpl::DocumentLayout QPdfViewImpl::calculateDocumentLayoutBook() const
{
	DocumentLayout documentLayout;

	if (!m_document || (m_document->status() != CorePDF::Document::Ready))
	{
		return documentLayout;
	}

	QHash<int, QRect> pageGeometries;

	const int pageCount = m_document->pageCount() - 1;
	const int curPage   = m_pageNavigation->currentPage();

	int totalWidth = 0;

	int startPage = 0;
	int endPage   = pageCount;

	if (not m_continuous)
	{
		/** First page */
		if (curPage == 0)
		{
			startPage = 0;
			endPage   = 0;
		}

		/** Odd pages */
		else if (curPage % 2 == 1)
		{
			startPage = curPage;
			endPage   = curPage;
		}

		/** Even pages */
		else
		{
			startPage = curPage - 1;
			endPage   = curPage - 1;
		}
	}

	const int horizMargins = m_documentMargins.left() + m_documentMargins.right();

	// calculate page sizes
	/** First page */
	if (startPage == 0)
	{
		QSizeF p1 = m_document->pageSize(0) * m_screenResolution;

		switch (m_renderOpts.rotation())
		{
		/* 90 degree rotated */
		case CorePDF::RenderOptions::Rotate90:
		case CorePDF::RenderOptions::Rotate270:
			/** Swap width <-> height */
			p1.transpose();

		default:
			break;
		}

		if (m_zoomMode == QPdfView::CustomZoom)
		{
			p1 *= m_zoomFactor;
		}

		else if (m_zoomMode == QPdfView::FitToWidth)
		{
			const qreal factor = (qreal(m_viewport.width() - horizMargins) / qreal(2 * p1.width() + m_pageSpacing));
			p1 *= factor;
		}

		else if (m_zoomMode == QPdfView::FitInView)
		{
			const int   margins = (m_continuous ? 0 : m_documentMargins.bottom() + m_documentMargins.top() - m_pageSpacing);
			const QSize viewportSize(m_viewport.size() - QSize(horizMargins, m_pageSpacing + margins));

			p1 = p1.scaled(viewportSize, Qt::KeepAspectRatio);
		}

		totalWidth        = qMax(totalWidth, p1.toSize().width() + horizMargins);
		pageGeometries[0] = QRect(QPoint(0, 0), p1.toSize());
	}

	for (int page = (startPage == 0 ? 1 : startPage); page <= endPage; page += 2)
	{
		QSizeF pageSize;

		QSizeF p1 = m_document->pageSize(page) * m_screenResolution;
		QSizeF p2(0, 0);

		if (page + 1 < m_document->pageCount())
		{
			p2 = m_document->pageSize(page + 1) * m_screenResolution;
		}

		switch (m_renderOpts.rotation())
		{
		/* 90 degree rotated */
		case CorePDF::RenderOptions::Rotate90:
		case CorePDF::RenderOptions::Rotate270:
			/** Swap width <-> height */
			p1.transpose();
			p2.transpose();

		default:
			break;
		}

		/** Page size is sum of individual pages, suitably pre-rotated */
		pageSize = QSizeF(p1.width() + p2.width(), qMax(p1.height(), p2.height()));

		if (m_zoomMode == QPdfView::CustomZoom)
		{
			pageSize *= m_zoomFactor;
			p1       *= m_zoomFactor;
			p2       *= m_zoomFactor;

			if (not not p2.isValid())
			{
				pageSize += QSizeF(m_pageSpacing, 0);
			}
		}

		else if (m_zoomMode == QPdfView::FitToWidth)
		{
			const qreal factor = qreal(m_viewport.width() - horizMargins - m_pageSpacing) / qreal(pageSize.width());
			pageSize *= factor;
			p1       *= (factor / (p2.width() <= 0 ? 2.0 : 1.0));
			p2       *= factor;
			if (not not p2.isValid())
			{
				pageSize += QSizeF(m_pageSpacing, 0);
			}
		}

		else if (m_zoomMode == QPdfView::FitInView)
		{
			const int   margins = (m_continuous ? 0 : m_documentMargins.bottom() + m_documentMargins.top() - m_pageSpacing);
			const QSize viewportSize(m_viewport.size() - QSize(horizMargins, m_pageSpacing + margins));

			pageSize = pageSize.scaled(viewportSize, Qt::KeepAspectRatio) + (not not p2.isValid() ? QSizeF(m_pageSpacing, 0) : QSizeF(0, 0));
			if (not p2.isValid())
			{
				p1 = QSizeF(pageSize.width(), pageSize.height());
			}

			else
			{
				p1        = QSizeF(pageSize.width() / 2.0, pageSize.height());
				p2        = QSizeF(pageSize.width() / 2.0, pageSize.height());
				pageSize += QSizeF(m_pageSpacing, 0);
			}
		}

		totalWidth = qMax(totalWidth, pageSize.toSize().width() + horizMargins);

		pageGeometries[page] = QRect(QPoint(0, 0), p1.toSize());
		if (not not p2.isValid())
		{
			pageGeometries[page + 1] = QRect(QPoint(0, 0), p2.toSize());
		}
	}

	int pageY = m_documentMargins.top();

	// calculate page positions
	if (startPage == 0)
	{
		const QSize s1         = pageGeometries[0].size();
		const int   pageHeight = s1.height();
		const int   pageX      = (qMax(totalWidth, m_viewport.width()) - s1.width()) / 2;
		pageGeometries[0].moveTopLeft(QPoint(pageX, pageY));
		pageY += pageHeight + m_pageSpacing;
	}

	for (int page = (startPage == 0 ? 1 : startPage); page <= endPage; page += 2)
	{
		const QSize size1 = pageGeometries.value(page).size();
		const QSize size2 = pageGeometries.value(page + 1).size();

		const int pageHeight = qMax(size1.height(), size2.height());

		// center horizontal inside the viewport
		const int pageX = (qMax(totalWidth, m_viewport.width()) - size1.width() - (size2.isNull() ? 0 : size2.width() + m_pageSpacing)) / 2;

		pageGeometries[page].moveTopLeft(QPoint(pageX, pageY));

		if (size2.isValid())
		{
			pageGeometries[page + 1].moveTopLeft(QPoint(pageX + size1.width() + m_pageSpacing, pageY));
		}

		pageY += pageHeight + m_pageSpacing;
	}

	pageY += m_documentMargins.bottom() - m_pageSpacing;

	documentLayout.pageGeometries = pageGeometries;

	// calculate overall document size
	documentLayout.documentSize = QSize(totalWidth, pageY);

	return documentLayout;
}


QPdfViewImpl::DocumentLayout QPdfViewImpl::calculateDocumentLayoutOverview() const
{
	DocumentLayout docLyt;

	return docLyt;
}


qreal QPdfViewImpl::yPositionForPage(int pageNumber) const
{
	const auto it = m_documentLayout.pageGeometries.constFind(pageNumber);

	if (it == m_documentLayout.pageGeometries.cend())
	{
		return 0.0;
	}

	return (*it).y();
}


void QPdfViewImpl::updateDocumentLayout()
{
	m_documentLayout = calculateDocumentLayout();

	updateScrollBars();
}


qreal QPdfViewImpl::zoomFactor() const
{
	int page = m_pageNavigation->currentPage();

	QSize pageSize;

	if (m_zoomMode == QPdfView::CustomZoom)
	{
		return m_zoomFactor;
	}

	else if (m_zoomMode == QPdfView::FitToWidth)
	{
		pageSize = QSizeF(m_document->pageSize(page) * m_screenResolution).toSize();
		const qreal factor = (qreal(m_viewport.width() - m_documentMargins.left() - m_documentMargins.right()) / qreal(pageSize.width()));
		return factor;
	}

	else if (m_zoomMode == QPdfView::FitInView)
	{
		const QSize viewportSize(m_viewport.size() + QSize(-m_documentMargins.left() - m_documentMargins.right(), -m_pageSpacing));

		pageSize = QSizeF(m_document->pageSize(page) * m_screenResolution).toSize();
		pageSize = pageSize.scaled(viewportSize, Qt::KeepAspectRatio);

		return pageSize.width() / m_document->pageSize(page).width();
	}

	return 1.0;
}


void QPdfViewImpl::paintSearchRects(int page, QImage& img)
{
	/** Search Rects */
	if (searchRects.contains(page))
	{
		QColor hPen   = qApp->palette().color(QPalette::Highlight);
		QColor hBrush = QColor(hPen);
		hBrush.setAlphaF(0.50);

		if (not m_documentLayout.pageGeometries.contains(page))
		{
			return;
		}

		/** Page rectangle */
		QRectF pgRect = m_documentLayout.pageGeometries[page];
		if (pgRect.isNull())
		{
			return;
		}

		/** The rectangles are too small. Buff them up. */
		double xPad = 0.0;
		double yPad = 0.0;

		/** Document Page Size */
		QSizeF dPageSize = m_document->pageSize(page);

		/** Rendered Page Size */
		QSizeF rPageSize = pgRect.size();

		switch (m_renderOpts.rotation())
		{
		case CorePDF::RenderOptions::Rotate90:
		case CorePDF::RenderOptions::Rotate270:
			dPageSize.transpose();

		default:
			break;
		}

		/** Layout not done. */
		if (not (rPageSize.width() and rPageSize.height()))
		{
			return;
		}

		QPainter painter(&img);

		painter.save();
		painter.setRenderHint(QPainter::Antialiasing);
		painter.setCompositionMode(QPainter::CompositionMode_Darken);
		painter.setPen(hPen);
		painter.setBrush(hBrush);

		double xZoom = rPageSize.width() / dPageSize.width();
		double yZoom = rPageSize.height() / dPageSize.height();

		for (QRectF rect: searchRects.value(page))
		{
			xPad = 2;
			yPad = 4;

			/** Transform the rectangle: handle rotations */
			qreal x = 0.0, y = 0.0, w = 0.0, h = 0.0;
			switch (m_renderOpts.rotation())
			{
			case CorePDF::RenderOptions::Rotate0:
				x = rect.x() * xZoom - xPad / 2.0;
				y = rect.y() * yZoom - yPad / 2.0;
				w = rect.width() * xZoom + 2 * xPad;
				h = rect.height() * yZoom + yPad;

				break;

			case CorePDF::RenderOptions::Rotate90:

				x = pgRect.width() - rect.y() * xZoom - rect.height() * xZoom - yPad / 2.0;
				y = rect.x() * yZoom - xPad / 2.0;
				w = rect.height() * xZoom + 2 * yPad;
				h = rect.width() * yZoom + xPad;

				break;

			case CorePDF::RenderOptions::Rotate180:

				x = pgRect.width() - rect.x() * xZoom - rect.width() * xZoom - xPad / 2.0;
				y = pgRect.height() - rect.y() * yZoom - rect.height() * yZoom - yPad / 2.0;
				w = rect.width() * xZoom + 2 * xPad;
				h = rect.height() * yZoom + yPad;

				break;

			case CorePDF::RenderOptions::Rotate270:

				x = rect.y() * xZoom - yPad / 2.0;
				y = pgRect.height() - rect.x() * yZoom - rect.width() * yZoom - xPad / 2.0;
				w = rect.height() * xZoom + 2 * yPad;
				h = rect.width() * yZoom + xPad;

				break;

			default:
				break;
			}

			/** Paint the rectangle */
			painter.drawRoundedRect(QRectF(QPointF(x, y), QSizeF(w, h)), 2.0, 2.0);
		}

		painter.restore();
		painter.end();
	}
}


QT_END_NAMESPACE
